from .cleverutils import timer
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions
import time

def disable_logging(**kwargs):
    """ Experimental: run selenium in silent mode """
    options = webdriver.ChromeOptions()
    options.headless = kwargs.get("headless")
    options.add_experimental_option("excludeSwitches", ["enable-logging"])
    return options

class Login_to:
    """
    A collection of common login/scraper/collector functions for a variety of
    websites.  Each receives a (CleverSession) object as its argument, typically comprising:

    .browser : a selenium webbrowswer object that has already been initialised
    .username : typically derived from CleverSession and keyring
    .password : typically a CleverSession @property based on keyring
    """
    @staticmethod
    @timer
    def github(self, browser=None, **kwargs):
        """ Use selenium and CleverSession credentials to login to Github """
        if browser is None:
                browser = webdriver.Chrome(options=disable_logging(**kwargs))
        browser.implicitly_wait(kwargs.get("wait") or 3)
        browser.get(self.url)
        browser.find_element_by_id("login_field").send_keys(self.username)
        browser.find_element_by_id("password").send_keys(self.password)
        browser.find_element_by_name("commit").click()
        if not hasattr(self, "browsers"):
            self.browsers = []
        self.browsers += [browser]

    @staticmethod
    @timer
    def twitter(self, browser=None, **kwargs):
        """ Use selenium and CleverSession credentials to login to Github """
        if browser is None:
                browser = webdriver.Chrome(options=disable_logging(**kwargs))
        browser.implicitly_wait(kwargs.get("wait") or 3)
        browser.get(self.url)
        browser.find_element_by_name("session[username_or_email]").send_keys(self.username)
        browser.find_element_by_name("session[password]").send_keys(self.password)
        span = browser.find_elements_by_tag_name("span")
        [x for x in span if x.text=="Log in"][0].click()
        if not hasattr(self, "browsers"):
            self.browsers = []
        self.browsers += [browser]

    @staticmethod
    @timer
    def office365(self, browser=None, **kwargs):
        """ Use selenium and CleverSession credentials to login to Office365 """
        if browser is None:
                browser = webdriver.Chrome(options=disable_logging(**kwargs))
        browser.implicitly_wait(kwargs.get("wait") or 3)
        browser.get(self.url)
        browser.find_element_by_id("i0116").send_keys(self.username)
        browser.find_element_by_id("idSIButton9").click()
        browser.find_element_by_id("i0118").send_keys(self.password)
        time.sleep(2)
        browser.find_element_by_id("idSIButton9").click()
        if not hasattr(self, "browsers"):
            self.browsers = []
        self.browsers += [browser]

    @timer
    @staticmethod
    def satchelone(self, browser=None, **kwargs):
        """ Use selenium and CleverSession credentials to login to SatchelOne
        """
        if browser is None:
            browser = webdriver.Chrome(options=disable_logging(**kwargs))
        # from satchelone_config import userid, pw
        browser.implicitly_wait(kwargs.get("wait") or 3)
        browser.get(self.url)
        main_window = browser.window_handles[0]
        span = browser.find_elements_by_tag_name("span")
        [x for x in span if x.text=="Sign in with Office 365"][0].click()
        popup_window = browser.window_handles[1]
        browser.switch_to.window(popup_window)
        Login_to.office365()
        browser.switch_to.window(main_window)
        print("\n ⓘ  Waiting for SatchelOne dashboard to appear...")
        while browser.current_url != 'https://www.satchelone.com/dashboard':
            continue
        print("\n ✓  OK we're in!\n")
        if not hasattr(self, "browsers"):
            self.browsers = []
        self.browsers += [browser]



from cleverutils import cleverutils, cleverwebutils, clevergui
from .cleversession import CleverSession
from .cleverwebutils import Login_to
from .clevergui import start_gui, button_menu, text_input, get_folder, progress_bar




from pprint import pprint
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions
import webbrowser
from pathlib import Path
from cleverdict import CleverDict
import pyperclip
import inspect
import json
import keyring
from .clevergui import start_gui, button_menu, text_input
from .cleverutils import to_json, timer
from .cleverwebutils import Login_to
import threading

class CleverSession(CleverDict):
    """
    A CleverDict sub-class(*) intended to handle selenium webbrowser sessions
    and common repeatable tasks such as selecting a target website using
    PySimpleGUI, retrieving login credentials using keyring, automated login,
    and scraper/collection tasks.

    Also uses predefined login and scraper functions notably from:

    login_to.py
    collect.py

    (*) CleverDict provides easy data handling and auto-save features.
    """
    index = CleverDict()
    choices = {"https://github.com/login": "Github",
               "https://twitter.com": "Twitter",
               "https://www.satchelone.com/login": "SatchelOne"}
    keyring_config_root = keyring.util.platform_.config_root()
    keyring_data_root = keyring.util.platform_.data_root()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        start_gui(redirect=kwargs.get("redirect"))
        self.check_and_prompt("url")
        if self.get("url"):
            self.account = CleverSession.choices[self.url]
            self.get_username()
        if not self.get("dirpath"):
            self.dirpath = Path().cwd()
        self.max_browsers = 5

    def get_username(self):
        """
        Loads (last modified) username from keyring. May not work on iOS.
        """
        try:
            self.username = keyring.get_credential(self.account, None).username
        except:
            print("\n  ⚠ .get_username() only supported on Windows OS.")
            print("\n     Trying creating .username manually first.")

    def check_and_prompt(self, *args):
        """
        Checks for existence/non-False value of an argument, and if required
        calls the specified method to prompt for a value.  Different methods
        may be needed for different input types e.g. file, folder, checkboxes, database read, REST API call, or regular input().

        args : attributes (as strings) to look for; try to use logical order.
        """
        buttons = {"url": "Please enter a link (URL) to the target website:",}
        text = {"username": "Please enter a username/login ID{}:",
                "password": "Please enter a password{}:"}
        choices = CleverSession.choices
        for attribute in args:
            prompt = buttons.get(attribute) or text.get(attribute) or f"Please enter a value for .{attribute} :"
            prompt = prompt.replace("{}", f" for your {self.account} account" if self.get("account") else "")
            if attribute == "password":
                self.check_and_prompt("url", "username")
                if not keyring.get_password(choices[self.url], self.username):
                    self.set_password(text_input(prompt))
            elif not self.get(attribute):
                if attribute in buttons:
                    value = button_menu(choices)
                else:  # Includes attributes in neither buttons nor text
                    value = text_input(prompt)
                if value:
                    self[attribute] = value

    @property
    def password(self):
        """ Retrieve password from keyring """
        return keyring.get_password(CleverSession.choices[self.url], self.username)

    def set_password(self, value):
        """ Set password in keyring """
        if value:
            keyring.set_password(CleverSession.choices[self.url], self.username, value)

    def delete_password(self):
        """
        Delete password AND username from keyring.
        .username remains in memory but .password was only ever an @property.
        """
        keyring.delete_password(CleverSession.choices[self.url], self.username)

    @timer
    def login_with_webbrowsers(self, browsers=None):
        """
        KWARGS:

        wait : Seconds for selenium to implicitly_wait
        browsers : int > number of browsers to run concurrently
        """
        self.check_and_prompt("url", "username", "password")
        if not hasattr(self, "browsers"):
            self.browsers = []
        dispatch = {"github.com": Login_to.github,
                    "twitter.com": Login_to.twitter,
                    "satchelone.com": Login_to.satchelone}
        browserThreads = []
        if browsers is None:
            browsers = self.get("max_browsers") or 1
        # Adjust if main browser is already be running:
        browsers = browsers - len(self.browsers)
        for n in  range(browsers):
            for website, func in dispatch.items():
                if website in self.url:
                    browserThread = threading.Thread(target=func, args=[self])
                    browserThreads.append(browserThread)
                    browserThread.start()
                    break
        for browserThread in browserThreads:
                browserThread.join()

    def save(self, name, value):
        """ Generic auto-save confirmation applied CleverDict """
        if "password" not in str(name).lower() and name != "browsers":
            print(f" ⓘ  {name} = {value} {type(value)}")

    def start(self):
        """ Shortcut/Alias for starting a webbrowser session and logging in """
        self.login_with_webbrowser()


# setattr(CleverSession, "login_with_webbrowsers", login_with_webbrowsers)
# self= CleverSession()
# self.start()




"""
A collection of high level functions, classes, and methods tailored to the author's current level and style of Python coding.
"""
import time
import json
from pathlib import Path
import inspect
from itertools import islice

def get_time(time_format="numeric"):
    """ Returns the local time in a predefined format """
    if time_format == "numeric":
        # 12 digit, all numeric.  Useful as a concise timestamp
        return time.strftime("%Y%m%d%H%M", time.localtime())

def to_json(self, never_save = False, **kwargs):
    """
    Additional CleverDict method to serialise its data to JSON.

    >>> setattr(CleverDict, "to_json", to_json)

    KWARGS
    never_save: Exclude field in CleverDict.never_save if True eg passwords
    file: Save to file if True or filepath

    * Will probably be incorporated into future versions of CleverDict,
    * at which point this function will be redundant.
    """
    # .get_aliases finds attributes created after __init__:
    fields_dict = {k: self.get(k) for k in self.get_aliases()}
    if never_save:
        fields_dict = {k: v for k,v in fields_dict.items() if k not in never_save}
    # JSON can't serialise Path objects, so convert to str
    for k,v in fields_dict.items():
        if isinstance(v, Path):
            fields_dict[k] = str(v)
    json_str = json.dumps(fields_dict, indent=4)
    path = kwargs.get("file")
    if path:
        path = Path(path)
        with path.open("w") as file:
            file.write(json_str)
        frame = inspect.currentframe().f_back.f_locals
        ids = [k for k, v in frame.items() if v is self]
        id = ids[0] if len(ids) == 1 else "/".join(ids)
        print(f"\n ⓘ  Saved '{id}' in JSON format to:\n    {path.absolute()}")
        print()
    return json_str

def timer(func):
    """
    Wrapper to start the clock, runs func(), then stop the clock. Simples.
    Designed to work as a decorator... just put @timer in the line above the
    original function.
    """
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        data = func(*args, **kwargs)
        print(f"\n ⏱  Function {func.__name__!r} took {round(time.perf_counter()-start,2)} seconds to complete.\n")
        return (data)
    return wrapper

def list_batches(data, **kwargs):
    """ Yields a sublist with batch_size values.

    kwargs:
    batch_size : Maximum size of any sublist returned; last sublist may be less.
    browsers: Number of browers, to run; Calculate batch_size accordingly

    x = list_batches(data)
    try:
        while True:
            sublist = next(x)
            do_stuff(sublist)
    except StopIteration:
        pass
    finally:
        del iterator
    """
    batch_size = kwargs.get("batch_size")
    if not batch_size:
        batch_size = len(data) // (kwargs.get("browsers") or 5)
        if len(data) % batch_size:  # positive remainder
            batch_size+=1
    it = iter(data)
    for i in range(0, len(data), batch_size):
        yield (x for x in islice(it, batch_size))

def dict_batches(data, batch_size=10):
    """ Yields a subdictionary with batch_size keys.

    x = dict_batches(data)
    try:
        while True:
            subdict = next(x)
            do_stuff(subdict)
    except StopIteration:
        pass
    finally:
        del iterator
    """
    it = iter(data)
    for i in range(0, len(data), batch_size):
        yield {k:data[k] for k in islice(it, batch_size)}




# Auto-generated by easyPyPI: https://github.com/PFython/easypypi
# Preserve current formatting to ensure easyPyPI compatibility.

from pathlib import Path
from setuptools import find_packages
from setuptools import setup

NAME = "cleverutils"
GITHUB_USERNAME = "Pfython"
VERSION = "0.20"
DESCRIPTION = "Some handy Python utilities and code snippets used repeatedly by the author and considered beginner to intermediate level of difficulty.  Published just in case they're of use to other Pythonistas somehwere, some time."
LICENSE = "MIT License"
AUTHOR = "Peter Fison"
EMAIL = "peter@southwestlondon.tv"
URL = "https://github.com/Pfython/cleverutils"
KEYWORDS = "cleverutils utils, cleverdict, cleversession, selenium, keyring, pysimplegui"
CLASSIFIERS = "Development Status :: 3 - Alpha, Intended Audience :: Developers, Operating System :: OS Independent, Programming Language :: Python :: 3.6, Programming Language :: Python :: 3.7, Programming Language :: Python :: 3.8, Programming Language :: Python :: 3.9, Topic :: Internet :: WWW/HTTP :: Browsers, Topic :: Internet :: WWW/HTTP :: Session, Topic :: Utilities"
REQUIREMENTS = "cleverdict, pysimplegui, selenium, keyring, pyperclip"


def comma_split(text: str):
    """
    Returns a list of strings after splitting original string by commas
    Applied to KEYWORDS, CLASSIFIERS, and REQUIREMENTS
    """
    if type(text) == list:
        return [x.strip() for x in text]
    return [x.strip() for x in text.split(",")]


if __name__ == "__main__":
    setup(
        name=NAME,
        packages=find_packages(),
        version=VERSION,
        license=LICENSE,
        description=DESCRIPTION,
        long_description=(Path(__file__).parent / "README.md").read_text(),
        long_description_content_type="text/markdown",
        author=AUTHOR,
        author_email=EMAIL,
        url=URL,
        download_url=f"{URL}/archive/{VERSION}.tar.gz",
        keywords=comma_split(KEYWORDS),
        install_requires=comma_split(REQUIREMENTS),
        classifiers=comma_split(CLASSIFIERS),
    )



"""
A collection of commonly high level functions based on PysSimpleGUI and tailored to the author's current level and style of Python coding.
"""
import PySimpleGUI as sg

sg_options = {"title": "CleverUtils", "keep_on_top": True, "icon": "../cleverutils.ico"}

def start_gui(*args, **kwargs):
    """
    Toggles between normal output and routing stdout/stderr to PySimpleGUI

    redirect: send (almost) all stdout/stderr to Debug Window & replace print()
    """
    if kwargs.get("redirect"):
        global print
        old_print = print
        print = sg.Print
        options = {"do_not_reroute_stdout": False, "keep_on_top": True}
        print(**options)
        print("Rerouting stdout/stderr to PySimpleGUI Debug Window...")
    # sg.change_look_and_feel("DarkAmber")
    # sg.change_look_and_feel("DarkGreen")
    sg.change_look_and_feel("Python")
    # Redirect stdout and stderr to Debug Window:
    sg.set_options(
        message_box_line_width=80,
        debug_win_size=(100, 30),
        icon = "cleverutils.ico",
        font = "calibri 12",
    )

def button_menu(choices: iter, prompt=None, **kwargs):
    """
    Presents a general purpose button selection menu using PySimpleGUI.
    Returns the text of the selected button (the 'event')
    """
    prompt = prompt or "Please select a website:"
    width = max([len(x) for x in choices]) +2
    layout=[[sg.Text(prompt, font="calibri 14 bold")]]
    layout.extend(
            [[sg.Button(button_text=url, size=(width,1), font="calibri 12")]
                for url in choices])
    layout.extend([[sg.Text("You can use Tab & Space to navigate", font="calibri 11 italic")]])
    title = kwargs.get("title") or "CleverUtils"
    del kwargs['title']
    window = sg.Window(title, layout, keep_on_top=True, element_justification="center")
    event, _ = window.read()
    window.close()
    return event

def text_input(prompt, **kwargs):
    """
    Presents a general purpose prompt for text input using PySimpleGUI.
    Returns the text entered, or None if closed with Cancel or X.
    """
    global sg_options
    kwargs.update(sg_options)
    if "password" in prompt.lower():
        kwargs.update({"password_char": "*"})
    kwargs['default_text'] = kwargs.get('default_text') or ""
    return sg.popup_get_text(prompt, **kwargs)

def get_folder(prompt, **kwargs):
    """
    Presents a general purpose prompt for folder selection using PySimpleGUI.
    Returns the selected folder, or None if closed with Cancel or X.
    """
    global sg_options
    kwargs.update(sg_options)
    return sg.popup_get_folder(prompt, default_path=kwargs.get("default_path"))

def progress_bar(prompt, **kwargs):
    """ Returns a PySimpleGUI window object that can be dynamically updated e.g.

    window['progress'].update(1, 5)
    window['progress_text'].update("some text")
    """
    global sg_options
    kwargs.update(sg_options)
    title = kwargs.get("title") or "CleverUtils"
    del kwargs['title']
    layout = [[sg.Text(prompt, key="progress_text")],
        [sg.ProgressBar(1, orientation=kwargs.get("orientation") or 'h', size=(kwargs.get("height") or 20, kwargs.get("width") or 20), key='progress')],
        ]
    window = sg.Window(title, layout, **kwargs).Finalize()
    return window




# Tests for cleverutils

from .cleverutils import *

class Test_Group_1:
    def test_something(self):
        """ Something should happen when you run something() """
        assert something() == something_else
